const constants = require('.././constants.json');
const alpha = 0.99
let log_channel = constants.topics.log_channel,
    metrics = {  }

let kafka = require('kafka-node'),
    Producer = kafka.Producer,
    client = new kafka.KafkaClient({
        kafkaHost: constants.network.external.kafka_host,
        autoConnect: true
    }),
    producer = new Producer(client)

/**
 * Function called to report metric data related to functions
 * @param {JSON} metric 
 */
function collectMetrics(metric) {
    /**
     * If metrics for a new function comes in, 
     * provision required structure for the function
     */
    if (!(metric.functionHash in metrics)) {
        metrics[metric.functionHash] = {
            shortterm: {
                coldstart: 0,
                coldstart_total_request: 0,
                warm_total_request: 0,
                warmstart: 0,
                worker_count: 0
            }
        }
    }
    if (metric.type === 'coldstart') {
        metrics[metric.functionHash].shortterm.coldstart += metric.value
        metrics[metric.functionHash].shortterm.coldstart_total_request += 1
    } else if (metric.type === 'warmstart') {
        metrics[metric.functionHash].shortterm.warmstart += metric.value
        metrics[metric.functionHash].shortterm.warm_total_request += 1
    } else if (metric.type === 'scale') {
        metrics[metric.functionHash].worker_count = metric.value
    }
    
}


/**
 * Run periodically to calculate average runtime metrics like coldstart and
 * warmstart latencies.
 * The module provides two granularities for metrics - shortterm and longterm
 * shortterm - realtime data at a granularity of 5s (set in dispatch_manager/lib.js)
 *             shortterm data is calculated using Simple Moving Average (SMA)
 * longterm - longterm data is held and averaged out over a period of time.
 *            longterm data is calculated using Expontential Moving Average (EMA)
 */
function broadcastMetrics() {
    
    if (Object.keys(metrics).length !== 0) {
        for (let [functionHash, metric] of Object.entries(metrics)) {
            
            if (metric.longterm === undefined) {
                metric.longterm = {
                        coldstart: 0,
                        warmstart: 0
                }
            }
            /**
             * Shortterm moving average
             */
            metric.shortterm.coldstart /= (metric.shortterm.coldstart_total_request != 0) ?
                metric.shortterm.coldstart_total_request : 1
            metric.shortterm.warmstart /= (metric.shortterm.warm_total_request != 0) ?
                metric.shortterm.warm_total_request : 1
            /**
             * Longterm exponential moving average
             */
            if (metric.shortterm.coldstart != 0)
                metric.longterm.coldstart = (metric.longterm.coldstart != 0)? metric.longterm.coldstart * alpha
                    + metric.shortterm.coldstart * (1 - alpha) : metric.shortterm.coldstart
            if (metric.shortterm.warmstart != 0)
                metric.longterm.warmstart = (metric.longterm.warmstart != 0)? metric.longterm.warmstart * alpha
                    + metric.shortterm.warmstart * (1 - alpha) : metric.shortterm.warmstart

            metric.timestamp = Date.now()
        }
        
        let log = [{
            topic: log_channel,
            messages: JSON.stringify({
                metrics
            }),
            partition: 0
        }]
        producer.send(log, () => { })
        
        for (let [functionHash, metric] of Object.entries(metrics)) {
            metric.shortterm = {
                coldstart: 0,
                coldstart_total_request: 0,
                warm_total_request: 0,
                warmstart: 0,
                worker_count: 0
            }
        }
    }
}

module.exports = {
    collectMetrics, broadcastMetrics
}